%{
/* atributy vyrazu */
typedef struct {
    char *sig; /* Signatura typu vyrazu */
    char *kod; /* Kod na vyhodnotenie vyrazu (instrukcie) */
} ATRV;
%}

    /* definicia prenasanych hodnot */
%union  {
    char* meno;
    char* txt;
    char* sig;
    unsigned int mod;
    int ihod;
    double dhod;
    char* kod;
    ATRV vhod;
}

/* definicia term. symbolov a hodnoty nimi prenasane */
%token          BOOLEAN  DOUBLE  INT  LONG  VOID
%token          CLASS  EXTENDS  CONSTRUCTOR IMPLEMENTS IMPORT
%token          PUBLIC PRIVATE PROTECTED
%token          STATIC FINAL ABSTRACT NATIVE
%token          NEW THIS SUPER RETURN IF ELSE WHILE
%token <meno>   ID
%token <txt>    STR_LITERAL
%token <ihod>   TRIEDA INT_LITERAL
%token <dhod>   DBL_LITERAL
%token          NULL_LITERAL
%token          TRUE_LITERAL
%token          FALSE_LITERAL
%token          LZZAT  PZZAT  LOZAT  POZAT  LHZAT  PHZAT
%token          CIARKA  BCIARKA  BODKA  ROVNE
/* priradenie priorit a asociativnosti operatorom */
%right      ROVNE
%left       AND OR
%nonassoc   LT LE GT GE EQ NE
%left       PLUS MINUS
%left       KRAT DEL
%right      BODKA

/* priradenie prenasanych hodnot neterm. symbolom */
%type <ihod> NadTrieda
%type <sig>  NepParametre Parametre TypVoid TypDat
%type <mod>  ModTriedy ModClena Mod1 Mod2
%type <vhod> Vyraz Primarny
%type <vhod> MetVolanie NepArgumenty Argumenty
%type <kod>  Prikaz Prikazy


%{  /* Udajove struktury pouzivane pri preklade */
typedef unsigned int uint;

/* Datovy clen */
typedef struct {
    uint pristup; /* Modifikatory pristupu */
    char *meno;
    char *signatura;
} dclen;

/* Metoda */
typedef struct {
    uint pristup;     /* Modifikatory pristupu */
    char *meno;
    char *signaturaP; /* Signatura parametrov */
    char *signaturaR; /* Signatura navratovej hodnoty */
    int velkostLP;    /* Velkost pamate pre lokalne premenne */
    char *kod;        /* Kod metody */
} met;

struct {
    char *menoTr;     /* Meno triedy uvedene v kode */
    char *javaTr;     /* Uplne meno triedy pre JVM */
    char *sigTr;      /* Signatura triedy */
    char *menoNadTr;  /* Meno nadtriedy */
    int ixntr;        /* Index na nadtriedu v tabulke tried */
    uint pristup;     /* Modifikatory pristupu */
    int pocDatClenov; /* Pocet datovych clenov */
    dclen* DatCleny;  /* Pole informacii o datovych clenoch */
    int pocMetod;     /* Pocet metod */
    met* Metody;      /* Pole informacii o metodach */
} triedy[20];

int ix_tr = -1; /* Index aktualne spracovavanej triedy */
int ix_m  = -1; /* Index aktualne spracovavanej metody */


/* Tabulka lokalnych premennych aktualne spracovavanej metody */
struct {
    char *meno; /* Meno premennej */
    char *sig;  /* Signatura typu */
    int index;  /* Pozicia premennej v pamati */
} LP[20];

int ixlp=0; /* Prva volna pozicia v tabulke */
int szlp=0; /* Velkost lokalnych premennych */


/* Modifikatory su ulozene pomcou bitovej masky */
#define MOD_PUBLIC    0x0001
#define MOD_PRIVATE   0x0002
#define MOD_PROTECTED 0x0004
#define MOD_STATIC    0x0008
#define MOD_FINAL     0x0010
#define MOD_NATIVE    0x0100
#define MOD_ABSTRACT  0x0400

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

FILE *vyst;
int ii, i1, i2;
char sigpom[80];
    /* prototypy */
int yyparse(void), yylex(void);
void ulozImp(char *,char *);
void ulozTr(char *,int,uint);
void ulozDC(char *,char *,uint);
void ulozMet(char *,char *,char *,uint);
void yyerror(char*), xxerror(char*, char*);
void vypisTriedy(int, char *);
int najdiDC(char *, int, int *);
int najdiM(char*, char*, int, int*);
int najdiT(char*);
void ulozLP(char *meno,char *sig);
void initLP(int mod);
void ukoncLP(void);
int najdiLP(char *meno);

/* vlozenie lexikalneho analyzatora */
#include "lex.yy.c"
%}

%%
Subor   : Importy Trieda
        | Subor  Trieda
        ;
Importy : /* e */
        | Importy  Import
        ;
Import  : IMPORT  ID  NATIVE  STR_LITERAL
            { ulozImp($2, $4); }
        ;
Trieda  : CLASS  ModTriedy ID  NadTrieda  ImpRozhrania
            { ulozTr($3, $4, $2); }
          LZZAT  Cleny  PZZAT
        ;
ModTriedy :        { $$ = 0; }
        | PUBLIC   { $$ = MOD_PUBLIC; }
        | ABSTRACT { $$ = MOD_ABSTRACT; }
        | FINAL    { $$ = MOD_FINAL; }
        ;
NadTrieda : /* e */
            { $$ = 0; }
        | EXTENDS  TRIEDA
            { $$ = $2; }
        ;
ImpRozhrania : /* e */
        | ImpRozhrania IMPLEMENTS TRIEDA
        ;
Cleny   : /* e */
        | Cleny DatovyClen
        | Cleny Metoda
        | Cleny error
        ;
DatovyClen : ModClena TypDat  ID  BCIARKA
            { ulozDC($3, $2, $1); }
        ;
ModClena : Mod1 Mod2     { $$ = $1 | $2; }
        ;
Mod1    :           { $$ = 0; }
        | PUBLIC    { $$ = MOD_PUBLIC; }
        | PROTECTED { $$ = MOD_PROTECTED; }
        | PRIVATE   { $$ = MOD_PRIVATE; }
        ;
Mod2    :           { $$ = 0; }
        | STATIC    { $$ = MOD_STATIC; }
        | ABSTRACT  { $$ = MOD_ABSTRACT; }
        | FINAL     { $$ = MOD_FINAL; }
        | STATIC FINAL  { $$ = MOD_STATIC | MOD_FINAL; }
        ;
Metoda  : ModClena TypDat ID
            { initLP($1 & MOD_STATIC); }
            LOZAT  NepParametre  POZAT
            { ulozMet($3, $6, $2, $1); }
            TeloMet
        | ModClena TypVoid ID
            { initLP($1 & MOD_STATIC); }
            LOZAT  NepParametre  POZAT
            { ulozMet($3, $6, $2, $1); }
            TeloMet
        | CONSTRUCTOR { initLP(0); } LOZAT NepParametre POZAT
            { ulozMet("<init>", $4, "V", MOD_PUBLIC); }
            TeloMet
        ;
NepParametre : /* e */
            { $$ = (char*) malloc(3); strcpy($$, "()"); }
        | Parametre
            { $$ = strcat($1, ")"); }
        ;
Parametre : TypDat  ID
            {
             ulozLP($2, $1);
             /* Zaciatok konstrukcie signatury parametrov */
             $$ = (char*) malloc(256);
             strcpy($$, "(");
             strcat($$, $1);
            }
        | Parametre  CIARKA  TypDat  ID
            { ulozLP($4, $3); $$ = strcat($1, $3); }
        ;
TypVoid : VOID
            { $$ = "V"; }
        ;
TypDat      : INT
            { $$ = "I"; }
        | DOUBLE
            { $$ = "D"; }
        | BOOLEAN
            { $$ = "Z"; }
        | TRIEDA
            { $$ = triedy[$1].sigTr; }
        | TRIEDA LHZAT PHZAT
            { $$ = sigpom; sprintf(sigpom, "[%s", triedy[$1].sigTr); }
        ;
TeloMet : BCIARKA
            { ukoncLP(); triedy[ix_tr].Metody[ix_m].kod = NULL; }
        | LZZAT  Prikazy  PZZAT
            { ukoncLP(); triedy[ix_tr].Metody[ix_m].kod = $2; }
        ;
Prikazy : /* e */ { $$ = (char*) malloc(6148); $$[0] = 0; }
        | Prikazy Prikaz
            {if ($2 != NULL) { $$ = strcat($1, $2); free($2); }
             else $$ = $1;
            }
        | Prikazy error
        ;
Prikaz  : TypDat ZozIdent BCIARKA
            { $$ = NULL; }
        | Vyraz BCIARKA
           {
            $$ = (char*) malloc(12 + strlen($1.kod));
            if ($1.sig[0] == 'V') sprintf($$, "%s", $1.kod);
            if ($1.sig[0] == 'L') sprintf($$, "%s\tpop\n", $1.kod);
            if ($1.sig[0] == 'I') sprintf($$, "%s\tpop\n", $1.kod);
            if ($1.sig[0] == 'Z') sprintf($$, "%s\tpop\n", $1.kod);
            if ($1.sig[0] == 'D') sprintf($$, "%s\tpop2\n", $1.kod);
            if ($1.sig[0] == 'J') sprintf($$, "%s\tpop2\n", $1.kod);
            free($1.kod);
           }
        | RETURN BCIARKA
            { $$ = (char*) malloc(10); sprintf($$, "\treturn\n"); }
        | BCIARKA
            { $$ = NULL; }
        | RETURN Vyraz BCIARKA
            { $$ = NULL; }
        | WHILE LOZAT Vyraz POZAT Prikaz
            { $$ = NULL; }
        | IF LOZAT Vyraz POZAT Prikaz ELSE Prikaz
            { $$ = NULL; }
        | IF LOZAT Vyraz POZAT Prikaz
            { $$ = NULL; }
        | LZZAT Prikazy PZZAT
            { $$ = $2; }
        ;
ZozIdent : ID                   { ulozLP($1, $<sig>0); }
        | ZozIdent CIARKA ID    { ulozLP($3, $<sig>0); }
        ;
Vyraz   : Primarny          { $$ = $1; }
        | Vyraz PLUS Vyraz  { $$.kod = NULL; }
        | Vyraz MINUS Vyraz { $$.kod = NULL; }
        | Vyraz KRAT Vyraz  { $$.kod = NULL; }
        | Vyraz DEL Vyraz   { $$.kod = NULL; }
        | Vyraz LT Vyraz    { $$.kod = NULL; }
        | Vyraz LE Vyraz    { $$.kod = NULL; }
        | Vyraz GT Vyraz    { $$.kod = NULL; }
        | Vyraz GE Vyraz    { $$.kod = NULL; }
        | Vyraz EQ Vyraz    { $$.kod = NULL; }
        | Vyraz NE Vyraz    { $$.kod = NULL; }
        | Vyraz OR Vyraz    { $$.kod = NULL; }
        | Vyraz AND Vyraz   { $$.kod = NULL; }
        ;
Primarny : ID
           {
            ii = najdiLP($1);
            $$.sig = LP[ii].sig;
            $$.kod = (char*) malloc(20);
            if ($$.sig[0] == 'L')
                sprintf($$.kod, "\taload\t%d\n", LP[ii].index);
            if ($$.sig[0] == 'I')
                sprintf($$.kod, "\tiload\t%d\n", LP[ii].index);
            if ($$.sig[0] == 'Z')
                sprintf($$.kod, "\tiload\t%d\n", LP[ii].index);
            if ($$.sig[0] == 'D')
                sprintf($$.kod, "\tdload\t%d\n", LP[ii].index);
           }
        | ID ROVNE Vyraz
           {
            ii = najdiLP($1);
            $$.sig = LP[ii].sig;
            $$.kod = (char*) malloc(20 + strlen($3.kod));
            if (strcmp(LP[ii].sig, $3.sig) != 0)
                yyerror("Nekompatibilne priradenie");
            if (LP[ii].sig[0] == 'L')
                sprintf($$.kod, "%s\tastore\t%d\n\taload\t%d\n", $3.kod, LP[ii].index, LP[ii].index);
            if (LP[ii].sig[0] == 'I')
                sprintf($$.kod, "%s\tistore\t%d\n\tiload\t%d\n", $3.kod, LP[ii].index, LP[ii].index);
            if (LP[ii].sig[0] == 'D')
                sprintf($$.kod, "%s\tdstore\t%d\n\tdload\t%d\n", $3.kod, LP[ii].index, LP[ii].index);
            if (LP[ii].sig[0] == 'Z')
                sprintf($$.kod, "%s\tistore\t%d\n\tiload\t%d\n", $3.kod, LP[ii].index, LP[ii].index);
            free($3.kod);
           }
        | DCPrirad      { $$.sig = "E"; $$.kod = NULL; }
        | THIS
           {
            $$.sig = triedy[ix_tr].sigTr;
            $$.kod = (char*) malloc(20);
            sprintf($$.kod, "\taload_0\n");
           }
        | NovyObjekt    { $$.sig = "E"; $$.kod = NULL; }
        | MetVolanie    { $$ = $1; }
        | DCPristup     { $$.sig = "E"; $$.kod = NULL; }
        | NULL_LITERAL
           {
            $$.sig = "A";
            $$.kod = (char*) malloc(20);
            sprintf($$.kod, "\taconst_null\n");
           }
        | INT_LITERAL
           {
            $$.sig = "I";
            $$.kod = (char*) malloc(20);
            sprintf($$.kod, "\tldc\t%d\n", $1);
           }
        | DBL_LITERAL
            {
             $$.sig = "D";
             $$.kod = (char*) malloc(20);
             sprintf($$.kod, "\tldc2_w\t%lf\n", $1);
            }
        | STR_LITERAL
           {
            $$.sig = triedy[1].sigTr;
            $$.kod = (char*) malloc(10 + strlen($1));
            sprintf($$.kod, "\tldc\t\"%s\"\n", $1);
           }
        | TRUE_LITERAL
           {
            $$.sig = "Z";
            $$.kod = (char*) malloc(12);
            sprintf($$.kod, "\ticonst_1\n");
           }
        | FALSE_LITERAL
           {
            $$.sig = "Z";
            $$.kod = (char*) malloc(12);
            sprintf($$.kod, "\ticonst_0\n");
           }
        | LOZAT  Vyraz  POZAT
            { $$ = $2; }
        | TRIEDA LOZAT Vyraz POZAT
           {
            $$.sig = triedy[$1].sigTr;
            $$.kod = $3.kod;
            if ($3.sig[0] != 'A')
                xxerror("Nemozno konvertovat na TRIEDU: ", $3.sig);
           }
        | INT LOZAT Vyraz POZAT
           {
            $$.sig = "I";
            if ($3.sig[0] == 'I') {
                $$.kod = $3.kod;
            } else if ($3.sig[0] == 'D') {
                $$.kod = (char*) malloc(10 + strlen($3.kod));
                sprintf($$.kod, "%s\td2i\n", $3.kod);
                free($3.kod);
            } else if ($3.sig[0] == 'J') {
                $$.kod = (char*) malloc(10 + strlen($3.kod));
                sprintf($$.kod, "%s\tl2i\n", $3.kod);
                free($3.kod);
            } else {
                xxerror("Nemozno konvertovat na INT: ", $3.sig);
                $$.kod = $3.kod;
            }
           }
        | DOUBLE LOZAT Vyraz POZAT
           {
            $$.sig = "D";
            if ($3.sig[0] == 'D') {
                $$.kod = $3.kod;
            } else if ($3.sig[0] == 'I') {
                $$.kod = (char*) malloc(10 + strlen($3.kod));
                sprintf($$.kod, "%s\ti2d\n", $3.kod);
                free($3.kod);
            } else if ($3.sig[0] == 'J') {
                $$.kod = (char*) malloc(10 + strlen($3.kod));
                sprintf($$.kod, "%s\tl2d\n", $3.kod);
                free($3.kod);
            } else {
                xxerror("Nemozno konvertovat na DOUBLE: ", $3.sig);
                $$.kod = $3.kod;
            }
           }
        | LONG LOZAT Vyraz POZAT
           {
            $$.sig = "J";
            if ($3.sig[0] == 'J') {
                $$.kod = $3.kod;
            } else if ($3.sig[0] == 'I') {
                $$.kod = (char*) malloc(10 + strlen($3.kod));
                sprintf($$.kod, "%s\ti2l\n", $3.kod);
                free($3.kod);
            } else if ($3.sig[0] == 'D') {
                $$.kod = (char*) malloc(10 + strlen($3.kod));
                sprintf($$.kod, "%s\td2l\n", $3.kod);
                free($3.kod);
            } else {
                xxerror("Nemozno konvertovat na LONG: ", $3.sig);
                $$.kod = $3.kod;
            }
           }
        ;
NovyObjekt : NEW  TRIEDA  LOZAT  NepArgumenty  POZAT
        ;
MetVolanie : TRIEDA BODKA ID LOZAT NepArgumenty POZAT
           {
            i2 = najdiM($3, $5.sig, $1, NULL);
            $$.sig = triedy[$1].Metody[i2].signaturaR;
            $$.kod = (char*) malloc(100 + strlen($5.kod));
            sprintf($$.kod, "%s\tinvokestatic\t%s/%s%s%s\n", $5.kod,
                triedy[$1].javaTr, $3, $5.sig, $$.sig);
            free($5.sig); free($5.kod);
           }
        | SUPER LOZAT NepArgumenty POZAT
           {
            i1 = triedy[ix_tr].ixntr;
            i2 = najdiM("<init>", $3.sig, i1, NULL);
            $$.sig = triedy[i1].Metody[i2].signaturaR;
            $$.kod = (char*) malloc(100 + strlen($3.kod));
            sprintf($$.kod, "\taload_0\n%s\tinvokespecial\t%s/<init>%s%s\n",
                $3.kod, triedy[i1].javaTr, $3.sig, $$.sig);
            free($3.sig); free($3.kod);
           }
        | Vyraz BODKA ID  LOZAT  NepArgumenty POZAT
           {
            i1 = najdiT($1.sig);i2 = najdiM($3, $5.sig, i1, &ii);
            $$.sig = triedy[ii].Metody[i2].signaturaR;
            $$.kod = (char*) malloc(100 + strlen($1.kod) + strlen($5.kod));
            sprintf($$.kod, "%s%s\tinvokevirtual\t%s/%s%s%s\n",
                $1.kod, $5.kod, triedy[i1].javaTr, $3, $5.sig, $$.sig);
            free($5.sig); free($5.kod);
           }
        | ID  LOZAT  NepArgumenty POZAT
           {
            i1  =  ix_tr; i2 = najdiM($1, $3.sig, i1, &ii);
            $$.sig = triedy[ii].Metody[i2].signaturaR;
            $$.kod = (char*) malloc(100 + strlen($3.kod));
            if (triedy[ii].Metody[i2].pristup & MOD_STATIC)
                sprintf($$.kod, "%s\tinvokestatic\t%s %s %s%s\n",
                    $3.kod, triedy[i1].javaTr, $1, $3.sig, $$.sig);
            else
                sprintf($$.kod, "\taload_0\n%s\tinvokevirtual\t%s/%s%s%s\n",
                    $3.kod, triedy[i1].javaTr, $1, $3.sig, $$.sig);
            free($3.sig); free($3.kod);
           }
        ;
DCPristup : TRIEDA BODKA ID
        | Vyraz BODKA ID
        ;
DCPrirad : TRIEDA BODKA ID  ROVNE Vyraz
        | Vyraz BODKA ID  ROVNE Vyraz
        ;
NepArgumenty : /* e */
           {
            $$.sig = (char*) malloc(4); strcpy($$.sig, "()");
            $$.kod = (char*) malloc(1); $$.kod[0] = 0;
           }
        | Argumenty
           { $$.sig = strcat($1.sig, ")"); $$.kod = $1.kod; }
        ;
Argumenty : Vyraz
           {
            $$.sig = (char*) malloc(256);
            strcpy($$.sig, "(");
            strcat($$.sig, $1.sig);
            $$.kod = (char*) malloc(4096);
            strcpy($$.kod, $1.kod);
           }
        | Argumenty  CIARKA  Vyraz
           {
            $$.sig = strcat($1.sig, $3.sig);
            $$.kod = strcat($1.kod, $3.kod);
           }
        ;
%%

static int pocetChyb = 0;

void yyerror(char *s)
{
    fprintf(stdout, "CHYBA: %s [%s:%d]\n", s, yytext, riadok);
    pocetChyb++;
}

void xxerror(char *s1, char *s2)
{
    fprintf(stdout, "CHYBA: %s %s [%d]\n", s1, s2, riadok);
    pocetChyb++;
}


int main(int argc, char *argv[])
{
    extern FILE *yyin;
    int i;
    char src[20]; /* Nazov vstupneho suboru */
    if (argc > 1) {
        yyin = fopen(argv[1], "r");
        strcpy(src, argv[1]);
        if (yyin == NULL) {
            yyin = stdin;
            strcpy(src, "stdin");
        }
    }
    ulozImp("Object", "java/lang/Object");
    ulozImp("String", "java/lang/String");
    ulozImp("System", "java/lang/System");
    (void) yyparse();
    fprintf(stderr, "POCET CHYB: %d\n", pocetChyb);
    for (i = 0; i <= ix_tr; i++)
        vypisTriedy(i, src);
    getchar();
    return 0;
}


/*
 * Nacitanie informacii o triede zo standardnej kniznice Java
 * meno - meno triedy
 * jmeno - uplne meno triedy
 * Informacie sa citaju zo suboru "java.api"
 */
void ulozImp(char *meno, char *jmeno)
{
    FILE *fi;
    char tok[80], pmeno[80]; /* token a precitane meno */
    uint pristup;
    fi = fopen("java.api", "r");
    if (fi == NULL) {
        xxerror("Nemozno najst archiv", "java.api");
        return;
    }
    fscanf(fi, "%s", tok);
    while (strcmp(tok, ".end") != 0) {
        if (strcmp(tok, ".class") != 0 && strcmp(tok, ".interface") != 0) {
            fscanf(fi, "%s", tok);
            continue;
        }
        fscanf(fi, "%x %s", &pristup, pmeno);
        if (strcmp(pmeno, jmeno) != 0) {
            fscanf(fi, "%s", tok);
            continue;
        }
        /* Zaciatok zaznamenavania triedy */
        ix_tr++;
        triedy[ix_tr].pristup = pristup | MOD_NATIVE;
        triedy[ix_tr].menoTr = meno;
        triedy[ix_tr].javaTr = jmeno;
        triedy[ix_tr].sigTr = (char*) malloc(strlen(jmeno)+3);
        sprintf(triedy[ix_tr].sigTr, "L%s;", jmeno);
        triedy[ix_tr].menoNadTr = NULL;
        triedy[ix_tr].ixntr = -1;
        fscanf(fi, "%s", tok);
        /* Nadtrieda */
        if (strcmp(tok, ".super") == 0) {
            fscanf(fi, "%s", pmeno);
            for (int i = 0; i < ix_tr; i++) {
                if (strcmp(triedy[i].menoTr, pmeno) == 0 ||
                        strcmp(triedy[i].javaTr, pmeno) == 0) {
                    triedy[ix_tr].menoNadTr = triedy[i].menoTr;
                    triedy[ix_tr].ixntr = i;
                }
            }
            if (triedy[ix_tr].ixntr < 0) {
                xxerror("Nemozno najst nadtriedu:", pmeno);
            }
        }
        /* Datove cleny */
        int pdc; /* pocet datovych clenov */
        fscanf(fi, "%d\n", &pdc);
        triedy[ix_tr].pocDatClenov = pdc;
        if (pdc > 0)
            triedy[ix_tr].DatCleny = (dclen*) calloc(pdc, sizeof(dclen));
        for (int i = 0; i < pdc; i++) {
            char sig[80];
            fscanf(fi, ".field %x %s %s\n", &pristup, pmeno, sig);
            triedy[ix_tr].DatCleny[i].meno = strdup(pmeno);
            triedy[ix_tr].DatCleny[i].signatura = strdup(sig);
            triedy[ix_tr].DatCleny[i].pristup = pristup;
        }
        /* Metody */
        int pm; /* pocet metod */
        fscanf(fi, "%d\n", &pm);
        triedy[ix_tr].pocMetod = pm;
        if (pm > 0)
            triedy[ix_tr].Metody = (met*) calloc(pm, sizeof(met));
        for (int i = 0; i < pm; i++) {
            char sigP[80], sigR[80];
            fscanf(fi, ".method %x %s %s %s\n", &pristup, pmeno, sigP, sigR);
            triedy[ix_tr].Metody[i].pristup=pristup;
            triedy[ix_tr].Metody[i].meno=strdup(pmeno);
            triedy[ix_tr].Metody[i].signaturaP=strdup(sigP);
            triedy[ix_tr].Metody[i].signaturaR=strdup(sigR);
        }
        fclose(fi);
        return;
    }
    fclose(fi);
    xxerror("Nemozno najst JAVA triedu:", jmeno);
}


/*
 * Ulozenie informacii o triede do tabulky tried
 * meno - meno triedy
 * ixNT - index nadtriedy (v tabulke tried)
 * mod  - modifikatory
 */
void ulozTr(char *meno, int ixNT, uint mod)
{
    int i, j, k;
    ix_tr++;
    triedy[ix_tr].menoTr = meno;
    triedy[ix_tr].javaTr = meno;
    triedy[ix_tr].sigTr = (char*) malloc(strlen(meno) + 3);
    sprintf(triedy[ix_tr].sigTr, "L%s;", meno);
    triedy[ix_tr].ixntr = ixNT;
    if (triedy[ixNT].pristup & MOD_FINAL)
        xxerror("Nemozno dedit z FINAL triedy", triedy[ixNT].menoTr);
    triedy[ix_tr].menoNadTr = triedy[ixNT].menoTr;
    triedy[ix_tr].pristup = mod;
    /* Alokacia pamate pre informacie o dat. clenoch a metodach */
    triedy[ix_tr].pocDatClenov = 0;
    triedy[ix_tr].DatCleny = (dclen*) calloc(20, sizeof(dclen));
    triedy[ix_tr].pocMetod = 0;
    triedy[ix_tr].Metody = (met*) calloc(20, sizeof(met));
}


/*
 * Ulozenie informacii o datovom clene triedy
 * meno - meno datoveho clena
 * sig  - signatura (typ)
 * mod  - modifikatory
 */
void ulozDC(char *meno, char *sig, uint mod)
{
    int i, poc;
    poc = triedy[ix_tr].pocDatClenov;
    /* Test, ci datovy clen s danym menom nie je uz deklarovany */
    for (i = 0; i < poc; i++) {
        if (strcmp(triedy[ix_tr].DatCleny[i].meno, meno) == 0) {
            xxerror("Viacnasobna deklaracia datoveho clena:", meno);
            return;
        }
    }
    triedy[ix_tr].DatCleny[poc].meno = meno;
    triedy[ix_tr].DatCleny[poc].pristup = mod;
    if (mod & MOD_ABSTRACT)
        xxerror("Datovy clen nemoze byt abstraktny:", meno);
    triedy[ix_tr].DatCleny[poc].signatura = sig;
    triedy[ix_tr].pocDatClenov++;
}


/*
 * Ulozenie informacii o metode
 * meno - meno metody
 * sigP - signatura parametrov
 * sigR - signatura navratovej hodnoty
 * mod  - modifikatory
 */
void ulozMet(char *meno, char *sigP, char *sigR, uint mod)
{
    int i;
    char *sig;
    ix_m = triedy[ix_tr].pocMetod;
    /* Kontrola jedinecnosti mena metody */
    for (i = 0; i < ix_m; i++) {
        if (strcmp(triedy[ix_tr].Metody[i].meno, meno) == 0 &&
                strcmp(triedy[ix_tr].Metody[i].signaturaP, sigP) == 0) {
            xxerror("Viacnasobna deklaracia metody:", meno);
            return;
        }
    }
    triedy[ix_tr].Metody[ix_m].meno = meno;
    triedy[ix_tr].Metody[ix_m].pristup = mod;
    triedy[ix_tr].Metody[ix_m].signaturaP = sigP;
    triedy[ix_tr].Metody[ix_m].signaturaR = sigR;
    triedy[ix_tr].Metody[ix_m].kod = NULL;
    triedy[ix_tr].pocMetod++;
}


/*
 * Najdi datovy clen
 * meno - meno datoveho clena
 * ixt  - index triedy
 * ixdt - [OUT] index triedy, v ktorej je datovy clen definovany
 *        (moze to byt nadtrieda)
 * Vrati index datoveho clena v triede s indexom ixdt
 * V pripade, ze ixdt je NULL, nadtriedy sa neprehliadavaju
 */
int najdiDC(char *meno, int ixt, int *ixdt)
{
    if (ixdt != NULL)
        *ixdt = ixt;
    for (int i = 0; i < triedy[ixt].pocDatClenov; i++) {
        if (strcmp(meno, triedy[ixt].DatCleny[i].meno) == 0)
            return i;
    }
    if (ixdt == NULL || triedy[ixt].ixntr < 0) {
        xxerror("Neexistujuci dat. clen:", meno);
        return -1;
    }
    return najdiDC(meno, triedy[ixt].ixntr, ixdt);
}


/*
 * Najdi metodu
 * meno - meno metody
 * sigp - signatura parametrov
 * ixt  - index triedy
 * ixdt - [OUT] index triedy, v ktorej je metoda definovana
 *        (moze to byt nadtrieda)
 * Vrati index metory v triede s indexom ixdt
 * V pripade, ze ixdt je NULL, nadtriedy sa neprehliadavaju
 */
int najdiM(char *meno, char *sigp, int ixt, int *ixdt)
{
    if (ixdt != NULL)
        *ixdt = ixt;
    for (int i = 0; i < triedy[ixt].pocMetod; i++) {
        if (strcmp(meno, triedy[ixt].Metody[i].meno) == 0
            && strcmp(sigp, triedy[ixt].Metody[i].signaturaP) == 0)
            return i;
    }
    if (ixdt == NULL || triedy[ixt].ixntr < 0) {
        xxerror("Neexistujuca metoda:", meno);
        return -1;
    }
    return najdiM(meno, sigp, triedy[ixt].ixntr, ixdt);
}


/*
 * Najdi triedu
 * sig - signatura triedy
 * Vrati index triedy v tabulke tried
 */
int najdiT(char *sig)
{
    for (int i = 0; i <= ix_tr; i++)
        if (strcmp(sig, triedy[i].sigTr) == 0)
            return i;
    xxerror("Neexistujuca trieda:", sig);
    return -1;
}


/*
 * Inicializuj pracu s lokalnymi premennymi metody
 * modstatic - je metoda staticka?
 * Pri spracovani sa pouzivaju globalne premenne LP, ixlp a szlp.
 */
void initLP(int modstatic)
{
    ixlp = 0;
    if (modstatic)
        szlp = 0; // staticka metoda - nema implicitne THIS
    else szlp = 1;   // virtualna metoda - ma implicitne THIS
}

/*
 * Uloz informacie o lokalnej premennej aktualne spracovavanej metody
 * meno - meno premennej
 * sig  - signatura typu
 */
void ulozLP(char *meno, char *sig)
{
    LP[ixlp].meno = meno;
    LP[ixlp].sig = sig;
    LP[ixlp].index = szlp;
    szlp++;
    if (sig[0] =='D')
        szlp++;
    ixlp++;
}

/*
 * Najdi index lokalnej premennej v tabulke
 * meno - meno premennej
 */
int najdiLP(char *meno)
{
    for (int i = 0; i < ixlp; i++) {
        if (strcmp(meno, LP[i].meno) == 0)
            return i;
    }
    xxerror("Neexistujuca lokalna premenna:", meno);
    return 0;
}

/*
 * Ukonci pracu s lokalnymi premennymi a zapis potrebne informacie
 */
void ukoncLP(void)
{
    triedy[ix_tr].Metody[ix_m].velkostLP = szlp;
}


/*
 * Vypis kodu implicitneho konstruktora triedy
 * ix   - index triedy
 * vyst - vystupny subor
 */
void implicitnyKonstruktor(int ix, FILE *vyst)
{
    for(int i = 0; i < triedy[ix].pocMetod; i++) {
        if(strcmp("<init>", triedy[ix].Metody[i].meno) == 0
            && strcmp("()", triedy[ix].Metody[i].signaturaP) == 0)
            return;
    }
    //autoamticke doplnenie
    int ixn = triedy[ix].ixntr;
    fprintf(vyst, "\n.method public  <init>()V\n");
    fprintf(vyst, "\t.limit stack 50\n");
    fprintf(vyst, "\t.limit locals 3\n");
    fprintf(vyst, "\taload_0\n");
    fprintf(vyst, "\tinvokespecial\t%s/<init>()V\n", triedy[ixn].javaTr);
    fprintf(vyst, "\treturn\n");
    fprintf(vyst, ".end method\n");
}


/*
 * Textova podoba modifikatorov
 * mod - modifikatory v podobe bitovej masky
 */
char *vratMod(uint mod)
{
    static char textmod[80];
    textmod[0] = '\0';
    if (mod & MOD_PUBLIC)    strcat(textmod, "public ");
    if (mod & MOD_PROTECTED) strcat(textmod, "protected ");
    if (mod & MOD_PRIVATE)   strcat(textmod, "private ");
    if (mod & MOD_STATIC)    strcat(textmod, "static ");
    if (mod & MOD_FINAL)     strcat(textmod, "final ");
    if (mod & MOD_ABSTRACT)  strcat(textmod, "abstract ");
    if (mod & MOD_NATIVE)    strcat(textmod, "native ");
    return textmod;
}


/*
 * Vypis prelozenej podoby triedy do Jal soboru
 * ix - index triedy v tabulke tried
 * src - nazov suboru so zdrojovym textom
 * Nazov vystupneho suboru je odvodeny na zaklade mena triedy.
 */
void vypisTriedy(int ix, char *src)
{
    int i, ixn;
    char mt[80]; /* Nazov suboru */
    FILE *vyst;
    if (triedy[ix].pristup & MOD_NATIVE)
        return;
    strcpy(mt, triedy[ix].menoTr);
    strcat(mt, ".jal");
    vyst = fopen(mt, "w");
    if (vyst == NULL)
        xxerror("Nemozno vytvorit subor:", mt);

    /* Vypis informacii o triede */
    fprintf(vyst, ".source %s\n", src);
    fprintf(vyst, ".class %s %s\n", vratMod(triedy[ix].pristup), triedy[ix].menoTr);
    ixn = triedy[ix].ixntr; /* nadtrieda */
    fprintf(vyst, ".super %s\n", triedy[ixn].javaTr);
    /* Vypis datovych clenov */
    for (i = 0; i < triedy[ix].pocDatClenov; i++) {
        fprintf(vyst, ".field %s %s %s\n",
            vratMod(triedy[ix].DatCleny[i].pristup),
            triedy[ix].DatCleny[i].meno,
            triedy[ix].DatCleny[i].signatura);
    }
    /* Vypis metod */
    implicitnyKonstruktor(ix, vyst);
    for (i = 0; i < triedy[ix].pocMetod; i++) {
        fprintf(vyst, "\n.method %s %s%s%s\n",
            vratMod(triedy[ix].Metody[i].pristup),
            triedy[ix].Metody[i].meno,
            triedy[ix].Metody[i].signaturaP,
            triedy[ix].Metody[i].signaturaR);
        if (triedy[ix].Metody[i].kod != NULL) {
            fprintf(vyst, "\t.limit stack 50\n");
            fprintf(vyst, "\t.limit locals %d\n", triedy[ix].Metody[i].velkostLP);
            fprintf(vyst, triedy[ix].Metody[i].kod);
        }
        fprintf(vyst, ".end method\n");
    }
    fclose(vyst);
}
